//
// Copyright (c) 2014, Peter A.M. van Hoof.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// Peter A.M. van Hoof
// Royal Observatory of Belgium
// Ringlaan 3
// B-1180 Brussels
// Belgium
// p.vanhoof@oma.be
//
//
// If you use any of these data in a scientific publication, please refer to
//
// van Hoof P.A.M., Williams R.J.R., Volk K., Chatzikos M., Ferland G.J., Lykins M., Porter R.L.
// Accurate determination of the free-free Gaunt factor, II -- relativistic Gaunt factors
// 2015, MNRAS, 449, 2112
//

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <string.h>

static const long gaunt_magic = 20140210L;
static const long gaunt_magic2 = 20140510L;

static const char* gauntff_file_nr  = "gauntff.dat";

#define NP_GAM2 81
#define NP_U 146

double gff_rel[NP_U][NP_GAM2];
double gff_nr[NP_U][NP_GAM2];

double lg_gam2_min, lg_u_min, step, lg_gam2_max, lg_u_max;

int min(int x, int y)
{
	return ( x < y ) ? x : y;
}

void read_table(const char* fnam, double gff[NP_U][NP_GAM2], bool lgPrint)
{
	char* ptr;
	long magic;
	size_t np_gam2, np_u, ipu, ipg2, len;
	FILE* io = fopen( fnam, "r" );
	if( io == NULL )
	{
		fprintf( stderr, "failed to open file %s\n", fnam );
		exit(1);
	}
	ptr = NULL;
	/* read magic number */
	while( true )
	{
		getline( &ptr, &len, io );
		if( lgPrint )
			printf( "%s", ptr );
		if( ptr[0] != '#' )
			break;
		free( ptr );
		ptr = NULL;
	}
	sscanf( ptr, "%ld", &magic );
	free( ptr );
	ptr = NULL;
	if( magic != gaunt_magic && magic != gaunt_magic2 )
	{
		fprintf( stderr, "read_table() found wrong magic number in file %s.\n", fnam );
		exit(1);
	}
	/* read dimensions of the table */
	getline( &ptr, &len, io );
	if( lgPrint )
		printf( "%s", ptr );
	sscanf( ptr, "%ld %ld",  &np_gam2, &np_u );
	free( ptr );
	ptr = NULL;
	assert( np_gam2 == NP_GAM2 && np_u == NP_U );
	/* read start value for log(gamma^2) */
	getline( &ptr, &len, io );
	if( lgPrint )
		printf( "%s", ptr );
	sscanf( ptr, "%le",  &lg_gam2_min );
	free( ptr );
	ptr = NULL;
	/* read start value for log(u) */
	getline( &ptr, &len, io );
	if( lgPrint )
		printf( "%s", ptr );
	sscanf( ptr, "%le",  &lg_u_min );
	free( ptr );
	ptr = NULL;
	/* read step size in dex */
	getline( &ptr, &len, io );
	if( lgPrint )
		printf( "%s", ptr );
	sscanf( ptr, "%le",  &step );
	free( ptr );
	ptr = NULL;
	lg_gam2_max = lg_gam2_min + (double)(np_gam2-1)*step;
	lg_u_max = lg_u_min + (double)(np_u-1)*step;
	/* read atomic number when present */
	if( magic == gaunt_magic2 )
	{
		getline( &ptr, &len, io );
		if( lgPrint )
			printf( "%s", ptr );
		free( ptr );
		ptr = NULL;
	}

	/* next lines are comments */
	for( int i=0; i < 9; ++i )
	{
		getline( &ptr, &len, io );
		if( i == 1 )
			strcpy( ptr+19, "the merged relativistic thermally averaged Gaunt factors.\n" );
		if( lgPrint )
			printf( "%s", ptr );
		free( ptr );
		ptr = NULL;
	}

	for( ipu=0; ipu < np_u; ++ipu )
	{
		for( ipg2=0; ipg2 < np_gam2; ++ipg2 )
			fscanf( io, "%le", &gff[ipu][ipg2] );
	}

	/* the remainder of the file contains the uncertainties of the gff values
	 * we will not read those here... */

	fclose(io);
}

int main(int, char** argv)
{
	read_table( argv[1], gff_rel, true );
	read_table( gauntff_file_nr, gff_nr, false );

	for( int ipu=0; ipu < NP_U; ++ipu )
	{
		int nborder = 3;
		int minpoint = 1000;
		double mindiff = 1e300;
		for( int ipg2=0; ipg2 < NP_GAM2; ++ipg2 ) {
			double diff = gff_rel[ipu][ipg2]/gff_nr[ipu][ipg2] - 1.;
			if( diff < 0. ) {
				minpoint = ( -diff < mindiff ) ? ipg2 : ipg2-1;
				break;
			}
			if( diff > mindiff ) {
				minpoint = ipg2-1;
				if( mindiff > 1.e-1 )
					nborder = 9;
				else if( mindiff > 3.e-2 )
					nborder = 7;
				else if( mindiff > 1.e-2 )
					nborder = 5;
				break;
			}
			mindiff = diff;
		}
		int hborder = nborder/2;
		for( int ipg2=0; ipg2 < min(minpoint-hborder,NP_GAM2); ++ipg2 ) {
			printf( "  %.8e", gff_rel[ipu][ipg2] );
		}
		for( int j=-hborder; j <= hborder; ++j ) {
			int ipg2 = minpoint+j;
			if( ipg2 < NP_GAM2 ) {
				double mid = sqrt(gff_rel[ipu][ipg2]*gff_nr[ipu][ipg2]);
				for( int k=1; k <= abs(j); ++k ) {
					if( j < 0 )
						mid = sqrt(mid*gff_rel[ipu][ipg2]);
					else
						mid = sqrt(mid*gff_nr[ipu][ipg2]);
				}
				if( j == hborder )
					mid = sqrt(sqrt(mid*mid*mid*gff_nr[ipu][ipg2]));
				printf( "  %.8e", mid );
			}
		}
		for( int ipg2=minpoint+hborder+1; ipg2 < NP_GAM2; ++ipg2 ) {
			printf( "  %.8e", gff_nr[ipu][ipg2] );
		}
		printf( "\n" );
	}
	return 0;
}
